# this is a decorator that will make a class a singleton
# usage:
# @singleton
# class MyClass:
#     pass
#
def singleton(cls):
    instances = {}

    def getinstance(*args, **kwargs):
        if cls not in instances or "recreate_singleton" in kwargs:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return getinstance



import glob
import json
import os
import shutil
from typing import Dict
import tomllib


class FileUtil:
    @staticmethod
    def load_to_string(path: str) -> str:
        with open(path, encoding="UTF-8") as user_file:
            file_contents = user_file.read()
        return file_contents

    @staticmethod
    def load_toml_to_dict(path: str) -> Dict:
        with open(path, 'rb') as toml_file:
            toml_dict = tomllib.load(toml_file)
        return toml_dict

    @staticmethod
    def load_json_to_dict(path: str) -> Dict:
        json_str = FileUtil.load_to_string(path)
        return json.loads(json_str)

    @staticmethod
    def save_to_json(path: str, data: Dict):
        json_object = json.dumps(data, indent=4)

        # Writing to sample.json
        with open(path, "w") as outfile:
            outfile.write(json_object)

    @staticmethod
    def save_to_toml(path: str, data: Dict):
        with open(path, "wb") as toml_file:
            tomllib.dump(data, toml_file)

    @staticmethod
    def move_files_with_prefix(file_pattern: str, target_folder: str, prefix: str = ""):
        """
        file_pattern = 'path/to/files/*.txt'
        target_folder = 'path/to/target/folder'
        prefix = 'new_'

        move_files_with_prefix(file_pattern, target_folder, prefix)
        """
        files = glob.glob(file_pattern)
        for file_path in files:
            file_name = os.path.basename(file_path)
            new_file_name = prefix + file_name
            new_file_path = os.path.join(target_folder, new_file_name)
            shutil.move(file_path, new_file_path)



import logging
import os
from typing import Optional

from config_profile.dict_util import DictUtil
from config_profile.file_util import FileUtil

logger = logging.getLogger(__name__)


class RequiredConfigKeyException(Exception):
    pass


class Config:
    """
    Gets a config value from various locations.
    The First location to supply a value wins.
    Values are searched for in this order:
        1) Pull from os env
        2) Pull from application-{env}.toml
        3) Pull from application.toml

    keys are lower case and use "." as a separator with the exception that
    not all os env vars support "." so for only env vars we also check for the uppercase key with "_" instead of "."
    (see unit tests for examples)

    Note: this class implements the singleton pattern
    """

    def __init__(self, resource_dir: str):
        self._config = {}

        # load values from the default application cfile
        self._populate_with_values_from_toml(config_file=os.path.join(resource_dir, "application.toml"))

        # load values from the env specific profile (ie local, dev, prod)
        profile = self.get_profile()
        self._populate_with_values_from_toml(config_file=os.path.join(resource_dir, f"application-{profile}.toml"))

    def get_dataset_name(self, base_name: str = "network_intel"):
        profile = self.get_profile()

        if profile in ["prod", "staging", "ci"]:
            return base_name
        elif profile in ["local"]:
            return f"{base_name}_qa"
        else:
            return f"{base_name}_{profile}"

    def _populate_with_values_from_toml(self, config_file: str):
        try:
            values = FileUtil.load_toml_to_dict(config_file)
            logger.info(f"Populating config with {config_file}")
            for k, v in DictUtil.flatten_dict(values, sep=".").items():
                self._config[k.lower()] = v
        except FileNotFoundError:
            if "application-local.toml" not in config_file:
                # also print because the logger is sometimes not initialized yet
                print(f"Config file not found: {config_file}")
                logger.warning(f"Config file not found: {config_file}")

    def get(self, key: str, default: Optional[str] = None) -> Optional[str]:
        # first check the os env
        v = self.get_env_value(key)
        if v:
            return v

        # check the application configs
        v = self._config.get(key)
        if v:
            return v
        return default

    def get_required(self, key: str) -> str:
        v = self.get(key)
        if v is None:
            raise RequiredConfigKeyException(f"Required key [{key}] not found in config")
        return v

    def is_enabled(self, key: str) -> bool:
        return self.get_boolean(key, default=False)

    def has_feature(self, feature_name: str) -> bool:
        return self.is_enabled(f"feature.{feature_name}.enabled")

    def get_boolean(self, key, default: Optional[bool] = None) -> Optional[bool]:
        v = self.get(key)
        if isinstance(v, bool):
            return v

        if v:
            return v.lower() == "true"
        return default

    def get_boolean_required(self, key) -> bool:
        v = self.get_required(key)
        return v.lower() == "true"

    def get_int(self, key, default: Optional[int] = None) -> Optional[int]:
        v = self.get(key)
        if v:
            return int(v)

        return default

    def get_int_required(self, key) -> int:
        v = self.get_required(key)
        return int(v)

    def get_float(self, key, default: Optional[float] = None) -> Optional[float]:
        v = self.get(key)
        if v:
            return float(v)

        return default

    def get_profile(self) -> str:
        # by design, the profile value can only be set from an os env
        profile = self.get_env_value("application.profile")

        if profile:
            return profile.lower()

        return "local"

    @classmethod
    def get_env_value(self, key: str, default: Optional[str] = None) -> Optional[str]:
        v = os.environ.get(key)
        if v:
            return v
        # several os envs vars do not support "." in the name, so we also check for the key with "_" instead of "."
        v = os.environ.get(key.upper().replace(".", "_"))
        if v:
            return v
        return default



from functools import reduce
from typing import Dict, List


class DictUtil:
    @staticmethod
    def merge(dicts: List[Dict]) -> Dict:
        """
        Merges dictionary recursively returning a new Dict
        """

        def _merge(a, b, path=None):
            # "merges b into a"
            if path is None:
                path = []
            for key in b:
                if key in a:
                    if isinstance(a[key], dict) and isinstance(b[key], dict):
                        _merge(a[key], b[key], path + [str(key)])
                    elif a[key] == b[key]:
                        pass  # same leaf value
                    else:
                        # raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
                        a[key] = b[key]
                else:
                    a[key] = b[key]
            return a

        return reduce(_merge, ([{}, *dicts]))

    @staticmethod
    def flatten_dict(d: dict, parent_key: str = "", sep: str = ".") -> dict:
        items = []
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, dict):
                items.extend(DictUtil.flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    @staticmethod
    def sort_keys(input_dict: dict) -> dict:
        """
        Sorts the keys of a dictionary and returns a new dictionary
        """
        # Get the keys from the dictionary and sort them
        sorted_keys = sorted(input_dict.keys())

        # Create a new dictionary with the sorted keys
        sorted_dict = {key: input_dict[key] for key in sorted_keys}

        return sorted_dict



import os
import pytest

from config_profile import Config, RequiredConfigKeyException

def test_base_config():
    config = Config(resource_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures", "resources")))

    assert config.get("top_level_key.some_sub_key") == "a value"

def test_config_profile_ordering():
    os.environ["APPLICATION_PROFILE"] = "test"
    config = Config(resource_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures", "resources")))

    assert config.get("top_level_key.key_to_be_overwritten") == "I have been overwritten"

def test_get():
    config = Config(resource_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures", "resources")))

    assert config.get("top_level_key.nonexistant_key", "default_value") == "default_value"

def test_get_required():
    config = Config(resource_dir=os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures", "resources")))

    assert config.get_required("top_level_key.some_sub_key") == "a value"
    with pytest.raises(RequiredConfigKeyException):
        config.get_required("top_level_key.nonexistant_key")
